class Bot extends Ship {
  Weapon weapon;
  
  protected float density = enemyShipDensity; //kg / m^2
  protected float friction = 0.3;
  protected float restitution = 0.50;
  protected final float epsilon = 0.00001;
  
  Vec2 steeringVector;
  boolean canPursue = false;
  int points = 1;
  
  boolean aboutToExplode = false;
  float explosionCountdown = 0;
  int explosionComboCount = 1;	//Used for combo tracking. Dead bots set this when they blow up, 
                                //and living ones check it when they're hit by that explosion.

  Bot(){
    this(new Vec2(0, 0), 1, 2);
  }
  
  Bot(Vec2 startPosition, float radius, float maxHitpoints){
    this.startPosition = startPosition.clone();
    this.radius = radius;
    this.maxHitpoints = maxHitpoints;
    this.hitpoints = maxHitpoints;
    
    steeringVector = new Vec2(0, 0);

    stickAroundTime = 5;
    maxThrust = 90000;
    maxSpeed = 40;
	
	enemiesStillAlive++;
  }
  
  Bot(Vec2 startPosition, float radius){
    this(
      startPosition, radius, 
      ceil(map(radius, 1, 4, 50, 200)/10)*10
    );
  }
  
  void addTo(World world){
    BodyDef bodydef = new BodyDef();
    bodydef.position.set(startPosition);
    startPosition = null; //we won't need it any more
    
    bodydef.angularDamping = 0.2;
    //bodydef.linearDamping = 0.1;
  
    CircleDef shapedef = new CircleDef();
    shapedef.radius = radius;
    shapedef.density = density;
    shapedef.friction = friction;
    shapedef.restitution = restitution;
  
    body = world.createBody(bodydef);
    body.createShape(shapedef);
    body.setMassFromShapes();
    body.setUserData(this);
  }
  
  void update(float dt){
    if ( dead ){
      return;
    }
	
	if ( aboutToExplode && !isTractored ){
		explosionCountdown -= dt;
		if ( explosionCountdown <= 0){
			die();
		}
	}
    
    if ( !isTractored ){
		steeringVector = calculateSteeringVector();
		//Limit the steering force to our maximum thrust
		if ( steeringVector.length() > maxThrust ){
			steeringVector.normalize();
			steeringVector.mulLocal(maxThrust);
		}
		body.applyForce(steeringVector, body.getWorldCenter());
      
		//Close enough to shoot?
		if ( weapon != null ){
			float targetDistance = playerShip.getPosition().sub(getPosition()).length();
			if ( targetDistance < getEngagementDistance() ) {
				weapon.fire();
			}
		}
    }
    
    if ( weapon != null ){
      weapon.update(dt);
    }
  }
  
  Vec2 calculateSteeringVector(){
	Vec2 steering = new Vec2(0, 0);
	
	if ( canPursue ){
		//"Intelligently" intercept the player
		steering.addLocal( pursue(playerShip) );
	} else {
		//Charge directly at the player
		steering.addLocal( seek(playerShip.getPosition()) );
	}
	
	//Avoid collisions with other bots
	steering.addLocal( separate(6, entities.getAll()).mul(2) );
	  
	return steering;
  }
  
  Vec2 steer(Vec2 target, float brakingDistance){
    float control = 0.10;
    float dt = 1.0 / 60.0; // The (average) physics step.

    Vec2 steer;
    
    Vec2 desiredVelocity = target.sub(getPosition());
    float distance = desiredVelocity.length();
    
    if (distance > 0) {
        // Compute the aiming direction
        desiredVelocity.normalize();
        
        if ( (brakingDistance == 0) || (distance > brakingDistance) ){
          // The ship's ideal velocity is the direction to the target multiplied by the max speed
          desiredVelocity.mulLocal(maxSpeed);
        } else {
          // Same as above, except the ships gradually slows down to a stop as it approaches the target
          desiredVelocity.mulLocal(maxSpeed * distance / brakingDistance);
        }        

        // Get the current ship's velocity because we will apply a force to compensate this.
        Vec2 currentVelocity = body.getLinearVelocity();

        // Compensate the current velocity by the desired velocity, based on the control factor
        Vec2 finalVelocity = desiredVelocity.sub(currentVelocity).mul(control);

        // Transform our velocity into an impulse (get rid of the time and mass factor)
        steer = finalVelocity.mul(body.getMass() / dt);
        
        if ( steer.length() > maxThrust ){
          steer.normalize();
          steer.mulLocal(maxThrust);
        }
    } else {
      steer = new Vec2(0, 0);
    }
    
    return steer;
  }
  
  Vec2 steer(Vec2 target){
    return steer(target, 0);
  }
  
  Vec2 seek(Vec2 target){
    return steer(target);
  }
  
  Vec2 flee(Vec2 target){
    return steer(target).negate();
  }
  
  Vec2 arrive(Vec2 target, float brakingDistance){
    return steer(target, brakingDistance);
  }
  
  Vec2 separateOld(float maxDistance, List<GameObject> flock){
    Vec2 accumulatedForce = new Vec2(0, 0); 
    float dt = 1.0 / 60.0;
    
    for(GameObject thing: flock){
      if (!(thing instanceof Bot)){
        continue;
      }
      Bot bot = (Bot)thing;
      Vec2 separationForce = getPosition().sub(bot.getPosition()); 
      
      //Check if the target is too far to matter
      float distance = separationForce.length() - radius - bot.radius;
      if ( distance > maxDistance ){
        continue;
      }
      //Steer away from the target
      separationForce.normalize();
      //The force is proportional to how close we are to the target
      Float r = max(distance, 0.0001) / maxDistance;
      separationForce.mulLocal( 1 / (r * r) );
      
      accumulatedForce.addLocal(separationForce);
    }
    
    //return steer(accumulatedForce.addLocal(getPosition()));
    //*
	//accumulatedForce.normalize();
    accumulatedForce.mulLocal(maxSpeed);
    accumulatedForce.mulLocal(body.getMass() / dt);

    //limitVector(accumulatedForce, maxThrust);
    return accumulatedForce;
  }
  
  Vec2 separate(float desiredSeparation, List<GameObject> flock, Class neighbourType){
	float dt = 1.0 / 60.0;
    Vec2 steer = new Vec2(0, 0); 
    int count = 0;
    
    for(GameObject thing: flock){
      if (!(neighbourType.isInstance(thing)) || (thing == this)){
        continue;
      }
	  assert (thing instanceof Ship) : "Can't execute separation behaviour against an entity that isn't a Ship";
	  
      Ship bot = (Ship)thing;
	  
      Vec2 diff = getPosition().sub(bot.getPosition()); 
	  float distance = diff.length() - radius - bot.radius;
	  
	  if ( (distance > 0) && (distance < desiredSeparation) ){
		diff.normalize();
		diff.mulLocal( 1f / (distance * distance) ); //or 1 / d^2
		steer.addLocal(diff);
		count++;
	  }
    }
	
	if ( steer.length() > 0 ){
		steer.normalize();
		steer.mulLocal(maxSpeed);
		steer.subLocal(body.getLinearVelocity());
		
        steer.mulLocal(body.getMass() / dt);
        
        if ( steer.length() > maxThrust ){
          steer.normalize();
          steer.mulLocal(maxThrust);
        }
	}
	
	return steer;
  }
  
  Vec2 separate(float desiredSeparation, List<GameObject> flock){
	return separate(desiredSeparation, flock, Bot.class);
  }
  
  // For every nearby ship in the system, calculate the average velocity
  Vec2 align (float neighbourDistance, List<GameObject> flock) {
	float dt = 1f / 60f;
    Vec2 steer = new Vec2(0, 0);
    int count = 0;
	
    for (GameObject thing : flock) {
		if ( !(thing instanceof Bot) || (thing == this) ){
			continue;
		}
		
		Bot bot = (Bot)thing;
		float distance = getPosition().sub(bot.getPosition()).length();
		
		if ( distance < neighbourDistance ){
			steer.addLocal(bot.body.getLinearVelocity());
			count++;
		}
    }
	
    if (count > 0) {
		steer.mulLocal(1 / (float)count);
    }

    // As long as the vector is greater than 0
    if (steer.length() > 0) {
		steer.normalize();
		steer.mulLocal(maxSpeed);
		steer.subLocal(body.getLinearVelocity());
		
		steer.mulLocal(body.getMass() / dt);
		
		if ( steer.length() > maxThrust ){
		  steer.normalize();
		  steer.mulLocal(maxThrust);
		}
    }
	
    return steer;
  }
  
  Vec2 pursue(Ship target){
	float distance = getPosition().sub(target.getPosition()).length();
	float lookAhead = distance / target.maxSpeed;
	Vec2 futureLocation = target.getPosition().add(target.body.getLinearVelocity().mul(lookAhead));
	return seek(futureLocation);
  }
  
  //Draw the bot ship
  void draw(){
    pushMatrix();
      Vec2 position = body.getPosition();
      translate(position.x, position.y);
      
      if ( !dead ){
        //drawHitpoints();
        drawSteeringVector();
      }

      //Display the body
      rotate(body.getAngle());
      drawHull();
      
      //Display the weapon
      if ( weapon != null && !dead ){
        weapon.draw();
      }
	  
	  //Some sort of visual marker for when we're about to explode
	  if ( !dead && aboutToExplode ){
		noFill();
		stroke(255, 20, 20);
		final float effectRadius = max(radius * 2.4, radius*2 + 1.5);
		ellipse(0, 0, effectRadius, effectRadius);
	  }
    popMatrix();
  }
  
  void drawHull(){
    if ( !dead ){
      stroke(0);
      fill(255);
    } else {
      stroke(180, getDeathlyAlpha());
      noFill();
    }
    
    ellipse(0, 0, radius*2, radius*2);
    rectMode(CENTER);
    rect(0, -radius/2, 1.0, radius);
    fill(255, 50, 50);
    
    //Display hitpoints as a growing bar
    if ( !dead ){
      noStroke();
      rect(0, -radius/2, 0.8, map(hitpoints, 0, maxHitpoints, radius, 0) );
    }
  }
  
  void drawHitpoints(){
    fill(0);
    textAlign(CENTER, BASELINE);
	textFont(font);
    
    pushMatrix();
      translate(0, - radius - 6.0/worldToScreenScale);
      scale(1.0/worldToScreenScale); //Necessary to prevent horrible text distortion
      textSize(24);
      text(ceil(hitpoints), 0, 0);
    popMatrix();
  }
  
  void drawSteeringVector(){
	final float visibleLength = radius + map(steeringVector.length()/maxThrust, 0, 1, 0, 2.5);
	Vec2 visibleVector = steeringVector.clone();
	visibleVector.normalize();
	visibleVector.mulLocal(visibleLength);
	
    stroke(50, 50, 255);
    line(0, 0, visibleVector.x, visibleVector.y);
  }
  
  float getDeathlyAlpha(){
    return map(
      constrain((remainingStickAroundTime / stickAroundTime), 0, 1),
      0, 1, 0, 255
    );
  }
  
  float getEngagementDistance(){
    return radius * 2 + shipRadius;
  }
  
  void die(float fatalDamage, GameObject damageDealer, int damageType) {
	if ( !aboutToExplode ){
		boolean killedByExplosion = 
			(damageType == EXPLOSIVE_DAMAGE) || ((damageDealer instanceof Bot) && ((Bot)damageDealer).aboutToExplode );
			
		if ( killedByExplosion ) {
			//Chain reaction. Big explosions lead to even bigger ones.
			if ( damageDealer instanceof Bot ){
				explosionComboCount += ((Bot)damageDealer).explosionComboCount;
			}
			
			lightTheFuse(1.0);
			return; //Delay death until the timed explosion or the next time the ship takes damage
			
		} else if ( damageDealer instanceof Bot ){
			explosionComboCount = 1;
		}
	}
	
	explode();
	
    //Dead ships don't collide with anything
    Shape sh = body.getShapeList();
    while(sh != null){
      sh.getFilterData().maskBits = 0;
      sh = sh.getNext();
    }
    
	//Add points
	awardPoints(points * explosionComboCount, explosionComboCount, getPosition());
	
    //This lets the Director know when to speed up the arrival of the next wave. Note: replace with observer pattern.
    enemiesStillAlive--;
	//Maybe reward the player with a powerup
	maybeSpawnPickup();
	
	super.die(fatalDamage, damageDealer, damageType);
  }
  
  void lightTheFuse(float countdown){
	aboutToExplode = true;
	explosionCountdown = countdown;
	hitpoints = 1; //Taking any more damage will cause the ship to go boom early.
  }
  
  void explode(){
	float multiplier = explosionComboCount;
	super.explode(multiplier);
  }
  
  void maybeSpawnPickup(){
	/*
	Vec2 powerupOffset = new Vec2(random(-radius*0.8, radius*0.8), random(-radius*0.8, radius*0.8));
	Pickup powerup = new PowerPickup( getPosition().addLocal(powerupOffset) );
	entities.add(powerup);
	//*/
  }
  
  void limitVector(Vec2 vector, float maxLength){
    if ( vector.length() > maxLength ){
        vector.normalize();
        vector.mulLocal(maxLength);
      }
  }
  
}

class BotContact {
  Bot bot;
  ContactPoint contact;
  
  BotContact(Bot bot, ContactPoint contact){
    this.bot = bot;
    this.contact = contact;
  }
}
