class PlayerShip extends Ship {
  Body hook;
  DistanceJoint beamJoint;
  
  boolean hookActive = false;
  Bot hookedBot = null;
  Joint hookJoint = null;
  
  float power = 180;
  float maxPower = 600;
  RepulsorBlast repulsor;

  PlayerShip(){
    super();
    hitpoints = maxHitpoints = playerHealth;
    radius = shipRadius;
    stickAroundTime = 6;
    
	repulsor = new RepulsorBlast(this);
	
	maxSpeed = 100;
  }
  
  void addTo(World world){
	//The player's ship
    BodyDef shipBodyDef = new BodyDef();
    shipBodyDef.position.set(0.0f, -10.0f);
    shipBodyDef.angularDamping = 0.9;
    shipBodyDef.linearDamping = 3;
    
    CircleDef shipShape = new CircleDef();
    shipShape.radius = shipRadius;
    shipShape.density = shipDensity;
    shipShape.friction = 0.7;
    
    body = world.createBody(shipBodyDef);
    body.createShape(shipShape);
    body.setMassFromShapes();
    body.setUserData(this);
    
    //The tractor beam "hook"
    BodyDef hookBodyDef = new BodyDef();
    hookBodyDef.position.x = shipBodyDef.position.x;
    hookBodyDef.position.y = shipBodyDef.position.y - shipRadius - hookRadius - initialBeamLength;
    hookBodyDef.linearDamping = 1;
    hookBodyDef.angularDamping = 0.9;
    hookBodyDef.fixedRotation = true;
    
    CircleDef hookShape = new CircleDef();
    hookShape.radius = hookRadius;
    hookShape.density = hookDensity;
    hookShape.isSensor = true; //pass through other bodies harmlessly
    
    hook = world.createBody(hookBodyDef);
    hook.createShape(hookShape);
    hook.setMassFromShapes();
    
    //Connect the hook to the ship
    DistanceJointDef def = new DistanceJointDef();
    def.initialize(body, hook, body.getWorldCenter(), hook.getWorldCenter());
    def.frequencyHz = 40.0;
    beamJoint = (DistanceJoint)world.createJoint(def);
	
	repulsor.addTo(world);
  }
  
  void draw(){
	//Give the repulsor weapon the chance to draw its GFX
	repulsor.draw();
  
    //Draw the tractor beam
    Vec2 position = playerShip.getPosition();
    Vec2 endPosition = hook.getPosition();
    stroke(dead ? 180 : 100);
    line(position.x, position.y, endPosition.x, endPosition.y);
    stroke(dead ? 180 : 0);
    fill(255);
    
    //Draw the ship
    pushMatrix();
      position = body.getPosition();
      translate(position.x, position.y);
      rotate(body.getAngle());
      stroke(dead ? 180 : 0);
      ellipse(0, 0, radius*2, radius*2);
      rect(0, -radius/2, 1, radius);
    popMatrix();
    
    //Draw the ship's tractor beam hook
    pushMatrix();
      position = hook.getPosition();
      translate(position.x, position.y);
      rotate(hook.getAngle());
      
	  if ( dead ){
		stroke(180);
	  } else if (hookActive) {
        stroke(255, 50, 50);
      } else {
        stroke(0);
      }
      
      ellipse(0, 0, hookRadius*2, hookRadius*2);
      rect(0, -hookRadius/2, 1, hookRadius);
    popMatrix();
  }
  
  void handleInput(){
	if ( dead ){
		return;
	}
	
	//Steer the ship with the mouse
	Vec2 force = new Vec2(mouseDx*mouseStrength, mouseDy*mouseStrength);
	body.applyForce(force, body.getWorldCenter()); //apply the force at the center of mass

    //Activate the tractor beam when LMB is down
    hookActive = mousePressed && (mouseButton == LEFT);
    
    //Release the tractored object with RMB
    if ( mousePressed && (mouseButton == RIGHT) ){
      if ( hookedBot != null ){
        releaseHookedObject();
      } else {
		/*
        //Debug thing: create an explosion 
        Explosion boom = new Explosion();
        boom.position = hook.getPosition();
        boom.radius = random(1.0, 3.0);
        boom.damage = 0;
        entities.add(boom);
		//*/
      }
    }
	
	//Fire the repulsor with SPACE
	if ( keyPressed && (key == ' ') && repulsor.readyToFire() ){
		repulsor.fire();
	}
  }  
  
  void update(float dt){
	if ( dead && (remainingStickAroundTime < 1.0) ){
		goToGameOver();
	}
  
    if ( hookedBot != null && (hookedBot.dead || this.dead)){
      releaseHookedObject();
    }
	
	if ( dead ){
		return;
	}
    
    //Check if the hook should latch onto one of the enemy ships
    if ( !possibleTargets.isEmpty() ){
      BotContact bc = possibleTargets.get(0); //Could be improved by sorting by separation, etc
      possibleTargets.clear();
          
      //Attach the hook to the target
      RevoluteJointDef def = new RevoluteJointDef();
      def.body1 = hook;
      def.localAnchor1 = hook.getLocalCenter();
      def.body2 = bc.bot.body;
      def.localAnchor2 = bc.bot.body.getLocalPoint(bc.contact.position);
      //def.initialize(hook, bc.bot.body, bc.contact.position);
      hookJoint = world.createJoint(def);
      hookedBot = bc.bot;
      bc.bot.isTractored = true;
    }
	
	repulsor.update(dt);
  }
  
  void releaseHookedObject(){
    world.destroyJoint(hookJoint);
    hookJoint = null;
    hookedBot.isTractored = false;
    hookedBot = null;
  }
  
  float getCollisionDamage(ContactResult result, PhysicalObject otherObject){
    if ( otherObject == null ){
      return 0;
    }
    
    //Take no damage from tractored ships
    if ( otherObject == hookedBot ){
      return 0;
    }
    
    return super.getCollisionDamage(result, otherObject);
  }
  
  void takeDamage(float amount, GameObject damageDealer, int damageType){
	//The player's ship is invulnerable to explosive damage
	if ( damageType == EXPLOSIVE_DAMAGE ){
		return;
	} else {
		super.takeDamage(amount, damageDealer, damageType);
	}
  }
  
  void heal(float amount){
	if ( !dead ){
		hitpoints = constrain(hitpoints + amount, 0, maxHitpoints);
	}
  }

  void gainPower(float amount){
	power = constrain(power + amount, 0, maxPower);
  }
 
  void deInit(){
	if ( hookedBot != null ){
		releaseHookedObject();
	}
	super.deInit();
  }
}
