import org.jbox2d.dynamics.contacts.*;
import org.jbox2d.collision.*;
import org.jbox2d.common.*;
import org.jbox2d.dynamics.joints.*;
import org.jbox2d.dynamics.*;

import java.awt.Robot;
import java.awt.AWTException;
import java.awt.event.InputEvent; 
import java.awt.MouseInfo;
import java.awt.Point;

int gameState = 0; //0 = intro, 1 = in game, 2 = game over
boolean paused = false;

World world;
Vec2 gravity = new Vec2(0.0f, 0.0f);
final float worldToScreenScale = 7;
float worldHalfWidth, worldHalfHeight;

Body[] walls;
final float wallThickness = 1.0;

PlayerShip playerShip;
final float shipRadius = 2.5; //meters
final float shipDensity = 450; //kg/m^2
final float mouseStrength = 8.2 * (PI * shipRadius * shipRadius * shipDensity);
final float playerHealth = 500;

final float hookRadius = 1.0f;
final float hookDensity = shipDensity / 3;

final float initialBeamLength = hookRadius * 3.5;
final float maxBeamLength = 15.0;
final float minBeamLength = 0.5;
final float beamExtensionSpeed = 0.4;
final float beamContractionSpeed = 0.4;

final float repulsorCost = 200;

final float enemyShipDensity = 100;

float score = 0;
int enemiesStillAlive = 0;

//Entities
EntityBag entities;
Director director;

//Physics
List<BotContact> possibleTargets;
MyContactListener contactHandler;

//Mouse control
Robot robot;
Point mouse, prevMouse;
float mouseDx, mouseDy;

PFont font, scoreFont;
ParticleEmitter collisionEmitter, textEmitter;

float gameTime = 0; //seconds
long lastTickTime = 0; //nanoseconds

void setup(){
  size(1024, 800);
  smooth();
  frameRate(60);
  fill(255);
  
  worldHalfWidth = width / 2 / worldToScreenScale;
  worldHalfHeight = height / 2 / worldToScreenScale;
  
  //Fonts
  font = loadFont("AgencyFB-Reg-48.vlw");
  textFont(font, 20);
  scoreFont = loadFont("AgencyFB-Bold-36.vlw");
  
  //Mouse
  try {
    robot = new Robot();
  } 
  catch (Exception e) {
    println("Can't Initialize the Robot");
  }
  addMouseWheelListener(new java.awt.event.MouseWheelListener() { 
    public void mouseWheelMoved(java.awt.event.MouseWheelEvent evt) { 
      mouseWheel(evt.getWheelRotation());
  }}); 
  
}

void draw(){
  switch(gameState){
    case 0: intro(); break;
	case 1: runGame(); break;
	case 2: gameOver(); break;
  }
}

void newGame(){
	gameState = 1;
	noCursor();
	prevMouse = MouseInfo.getPointerInfo().getLocation(); ;

	possibleTargets = new ArrayList<BotContact>();
	initPhysics();

	//Create entities
	entities = new EntityBag();

	director = new Director();
	entities.add(director);
	
	collisionEmitter = new ParticleEmitter(new CollisionParticleFactory());
	collisionEmitter.pVelocity = new FloatRange(40, 50);
	collisionEmitter.pLifetime = new FloatRange(0.2, 0.6);
	collisionEmitter.pDamping = new FloatRange(0.05, 0.3);
	collisionEmitter.rate = 0;
	entities.add(collisionEmitter);
	
	textEmitter = new ParticleEmitter(new TextParticleFactory());
	textEmitter.pVelocity = new FloatRange(1, 3);
	textEmitter.pLifetime = new FloatRange(2, 5);
	textEmitter.pDamping = new FloatRange(0.05, 0.1);
	textEmitter.pDirection = new VectorRange(new Vec2(0, -1), 0);
	textEmitter.rate = 0;
	entities.add(textEmitter);
	
	playerShip = new PlayerShip();
	entities.add(playerShip);
	entities.add( new HealthBar(playerShip) );
	entities.add( new PowerBar(playerShip) );
	
	paused = false;
	gameTime = 0;
	lastTickTime = System.nanoTime();
}

void runGame(){
  final long thisTickTime = System.nanoTime();
  final float dt = (thisTickTime - lastTickTime) / 1000000000f;
  if ( dt <= 0 ){
	return;
  }
  lastTickTime = thisTickTime;
  
  if ( paused ){
	return;
  }
  gameTime += dt;
  
  handleInput();
  entities.update(dt);
  updatePhysics(dt);
  
  background(255);
  drawWorld();
  drawUI();
}

void goToGameOver(){
	cursor();
	gameState = 2;
}

void gameOver(){
	background(255);
	fill(0);
	noStroke();
	textAlign(CENTER, BASELINE);
	
	pushMatrix();
	translate(width/2, height/2 - 50);
	
	textSize(50);
	text("GAME OVER", 0, 0);
	textSize(24);
	text("SCORE: " + ceil(score), 0, 32);
	popMatrix();
	
	fill(100);
	text("Press SPACE to start a new game or ESC to exit", width/2, height-20);
}

void goToIntro(){
	cursor();
	gameState = 0;
}

void intro(){
	background(255);
	fill(0);
	noStroke();
	textAlign(CENTER, BASELINE);
	
	pushMatrix();
		translate(width/2, height/2 - 50);
		textSize(50);
		text("Sling Prototype", 0, 0);
	popMatrix();
	
	fill(100);
	textSize(24);
	text("Press SPACE to start the game", width/2, height-20);
}

void keyReleased(){
	if ( key != ' ' ){
		return;
	}

	switch(gameState){
		case 0: //Intro => Play
			newGame();
			break;
		case 2: //Game Over => Intro
			goToIntro();
			break;
	}
}

void keyPressed(){
	if ( gameState == 1 ){
		if ( (key == 'p') || (key == 'P') ){
			paused = !paused;
			if ( paused ){
				cursor();
			} else {
				noCursor();
			}
		}
	}
}

void awardPoints(int amount, int comboLevel, Vec2 position){
	score += amount;
	
	if ( (playerShip != null) && !playerShip.dead ){
		playerShip.gainPower(amount / 10);
	}
	
	if ( (position != null) && (amount > 0) ){
		TextParticle particle = (TextParticle)(textEmitter.emitParticle());
		particle.content = "+" + Integer.toString(amount);
		particle.size = constrain(ceil(map(sqrt(amount), 3, 15, 18, 48)), 18, 64);
		particle.lifetime = map(sqrt(amount), 2, 15, 1, 7);
		particle.position.set(position);
	}
}